@@ -1,10 +1,16 @@
 import copy
+import json
+from decimal import Decimal
+
 import boto3
 
 from boto3.dynamodb.types import TypeSerializer, TypeDeserializer
 from botocore import exceptions
 
 
+class InsufficientArgumentsException(Exception):
+    pass
+
 class IndexNotValidException(Exception):
     pass
 
@@ -17,12 +23,12 @@ class DynamoDBDataStore(object):
     serializer = TypeSerializer()
     deserializer = TypeDeserializer()
 
-    def __init__(self, table_name):
+    def __init__(self, table_name, hash_key='PK', sort_key='SK', use_default_index_keys=True):
 
         DB_SETTINGS = {
             'TableName': table_name,
-            'HashKeyName': 'PK',
-            'SortKeyName': 'SK'
+            'HashKeyName': hash_key,
+            'SortKeyName': sort_key
         }
 
         self.table_name = DB_SETTINGS['TableName']
@@ -30,15 +36,33 @@ class DynamoDBDataStore(object):
         self.sort_key_name = DB_SETTINGS.get('SortKeyName', None)
         self.index_keys = [
             self.hash_key_name,
-            self.sort_key_name,
-            'GSI1PK',
-            'GSI1SK',
-            'GSI2PK',
-            'GSI2SK',
+            self.sort_key_name
         ]
+        if use_default_index_keys:
+            self.add_index_keys([
+                'GSI1PK',
+                'GSI1SK',
+                'GSI2PK',
+                'GSI2SK',
+            ])
+
         self.client = boto3.client('dynamodb')
 
-    def save_document(self, document, index=None, parameters={}):
+    def add_index_keys(self, keys_to_add:list):
+        if type(keys_to_add) != list:
+            raise TypeError('Expected a list but received ', type(keys_to_add))
+        for key in keys_to_add:
+            self.add_index_key(key)
+
+    def add_index_key(self, key_to_add: str):
+        if type(key_to_add) != str:
+            raise TypeError('Expected a string but received ', type(key_to_add))
+        self.index_keys.append(key_to_add)
+
+    def save_document(self, document, index=None, parameters=None):
+        if parameters is None:
+            parameters = {}
+
         document = copy.deepcopy(document)
 
         if len(index) not in [1, 2]:
@@ -50,22 +74,32 @@ class DynamoDBDataStore(object):
             self._save_document_using_composite_key(document, hash_key=index[0], sort_key=index[1],
                                                     parameters=parameters)
 
-    def _save_document_using_composite_key(self, document, hash_key, sort_key, parameters={}):
+    def _save_document_using_composite_key(self, document, hash_key, sort_key, parameters=None):
+        if parameters is None:
+            parameters = {}
         document[self.hash_key_name] = hash_key
         document[self.sort_key_name] = sort_key
 
+        document = json.loads(json.dumps(document), parse_float=Decimal)
+
         document = self.serialize(document)
         self._put_item(document, parameters=parameters)
 
-    def _save_document_using_hash_key(self, document, hash_key, parameters={}):
+    def _save_document_using_hash_key(self, document, hash_key, parameters=None):
+        if parameters is None:
+            parameters = {}
         document[self.hash_key_name] = hash_key
 
+        document = json.loads(json.dumps(document), parse_float=Decimal)
+
         document = self.serialize(document)
 
         self._put_item(document, parameters=parameters)
 
-    def _put_item(self, item, parameters={}):
+    def _put_item(self, item, parameters=None):
 
+        if parameters is None:
+            parameters = {}
         params = {
             'TableName': self.table_name,
             'Item': item,
@@ -119,8 +153,10 @@ class DynamoDBDataStore(object):
         paginator = self.client.get_paginator('query')
         return paginator.paginate(**parameters)
 
-    def update_document(self, index=None, parameters={}):
+    def update_document(self, index=None, parameters=None):
 
+        if parameters is None:
+            parameters = {}
         if len(index) not in [1, 2]:
             raise IndexNotValidException
 
@@ -144,7 +180,9 @@ class DynamoDBDataStore(object):
                 raise e
             raise ConditionalCheckFailedException
 
-    def delete_document(self, index=None, parameters={}):
+    def delete_document(self, index=None, parameters=None):
+        if parameters is None:
+            parameters = {}
         if len(index) not in [1, 2]:
             raise IndexNotValidException
 
@@ -169,23 +207,38 @@ class DynamoDBDataStore(object):
             raise ConditionalCheckFailedException
 
     def batch_request(self, request_items):
+        if not request_items:
+            raise InsufficientArgumentsException('You must provide request_items')
+
         self.client.batch_write_item(RequestItems={
             self.table_name: request_items
         })
 
-    def get_documents(self, query, return_index=False):
-        pages = self.paginate(query)
-        documents = []
+    def get_documents(self, query=None, return_index=False):
+        if not query:
+            raise InsufficientArgumentsException('You must provide a query')
+        try:
+            pages = self.paginate(query)
+            documents = []
+    
+            for page in pages:
+                items = page['Items']
+    
+                for item in items:
+                    document = self.deserialize(item)
+                    if not return_index:
+                        for key in self.index_keys:
+                            document.pop(key, None)
+    
+                    documents.append(document)
+    
+            return documents
+        except Exception as e:
+            print(e)
+            return None
 
-        for page in pages:
-            items = page['Items']
-
-            for item in items:
-                document = self.deserialize(item)
-                if not return_index:
-                    for key in self.index_keys:
-                        document.pop(key, None)
-
-                documents.append(document)
-
-        return documents
+    def transaction_write(self, batch_items, transaction_id):
+        self.client.transact_write_items(
+            TransactItems=batch_items,
+            ClientRequestToken=transaction_id
+        )
